
#pragma once

#include "stdafx.h"
#include "sample.h"
#include "sampleDlg.h"

#include "DIBSectn.h"

#include <mtype.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

volatile int g_save_enabled = 0;
volatile int g_frame_number = 0;
volatile int g_frames_file_number = 0;
volatile bool g_save_frame_format_info = false;
volatile bool g_save_scan_lines = false;
volatile bool g_conversion_in_progress = false;
volatile bool g_stop_conversion = false;
volatile bool g_close_in_progress = false;
LONGLONG g_frame_time;
LONGLONG g_frame_time_old;
LONGLONG g_frame_time_diff;
CString g_fn_str;

//--------------------------------------------------------------------------

void CsampleDlg::CreateUsgControl( IUsgDataView* data_view, const IID& type_id, ULONG scan_mode, ULONG stream_id, void** ctrl )
{
	IUsgControl* ctrl2;
	ctrl2 = NULL;
	if (data_view == NULL) return;
	data_view->GetControlObj( &type_id, scan_mode, stream_id, &ctrl2 );
	if (ctrl2 != NULL)
	{
		HRESULT hr;
		hr = ctrl2->QueryInterface( type_id, (void**)ctrl );
		if (hr != S_OK)
			*ctrl = NULL;
		SAFE_RELEASE(ctrl2);
	}
}

//--------------------------------------------------------------------------

void CsampleDlg::CreateUsgControls()
{
	HRESULT hr = S_OK;
	IUnknown* tmp_obj = NULL;
	g_frames_file_number = 0;

	do
	{
		CString fn_str;
		m_file_ctrl.GetWindowTextW(fn_str);
		CFileStatus status;
		if (CFile::GetStatus(fn_str, status) == FALSE)
		{
			MessageBox(L"File not found.",L"");
			break;
		}

		// create main Usgfw2 library object
		if (m_usgfw2 == NULL)
			hr = CoCreateInstance(CLSID_Usgfw2, NULL, CLSCTX_INPROC_SERVER, IID_IUsgfw2,(LPVOID*) &m_usgfw2);
		if ( (hr != S_OK) || (m_usgfw2 == NULL) )
		{
			m_usgfw2 = NULL;
			break;
		}

		hr = m_usgfw2->CreateUsgfwFile(CComBSTR(fn_str), FCF_OPEN, &m_storage);
		if (hr != S_OK)
		{
			m_storage = NULL;
			MessageBox(L"Cannot open file.",L"");
			break;
		}

		g_fn_str = fn_str;

		// create main ultrasound scanning object (data view)
		m_usgfw2->CreateDataView(m_storage, &m_data_view);
		if (m_data_view == NULL)
			break;

		// stop ultrasound scanning
		m_data_view->put_ScanState(SCAN_STATE_STOP);

		IUsgScanMode* mode = NULL;
		m_data_view->GetScanModeObj(SCAN_MODE_B,&mode);
		if (mode == NULL)
		{
			MessageBox(L"The file does not contain B mode data.",L"");
			break;
		}
		// get mixer control for B mode
		mode->GetMixerControl(SCAN_MODE_B,0,&m_mixer_control);
		SAFE_RELEASE(mode);
		if (m_mixer_control == NULL)
			break;

		// set B scanning mode
		m_data_view->put_ScanMode(SCAN_MODE_B);

		// set ultrasound output window;
		m_mixer_control->SetOutputWindow((LONG)(this->GetSafeHwnd()));
		
		tagRECT rect1;
		rect1.left		= 10;
		rect1.top		= 10;
		rect1.right		= rect1.left + 512;
		rect1.bottom	= rect1.top + 512;
		// set ultrasound output rectangle; output rectangle can have any size;
		// scan-converter will automatically create ultrasound image using correct aspect ratio;
		// selected scanning depth will be fit to the height of output rectangle;
		// ultrasound image width will be adjusted proportionally and, if required, the image will be cropped;
		// the image of requested size is calculated NOT by resizing some fixed output image, but
		// during scan-conversion and interpolation process when from scan-lines is created
		// 2D image using information about ultrasound probe geometry;
		// the number of points in one B mode ultrasound beam (scan-line) varies from ~700 to ~1024.
		// the number of beams (scan lines) in one B mode frame may be between 64 and 128
		// depending on the used ultrasound scanner, probe, lines density;
		m_mixer_control->SetOutputRect(&rect1);

		tagPALETTEENTRY clr1;
		clr1.peRed		= 0;
		clr1.peGreen	= 0;
		clr1.peBlue		= 0;
		clr1.peFlags	= 0;
		// set background color that surrounds ultrasound image
		m_mixer_control->put_BkColor(clr1);

		CreateUsgControl( m_data_view, IID_IUsgCineStream, SCAN_MODE_B, 0, (void**)&m_cine );
		if (m_cine == NULL)
			break;

		// create scan-converter callback COM object that can be implemented as class like in
		// this sample or can be implemented as a usual COM object that is created using CoCreateInstance
		m_scan_converter_callback = new ScanConverterCallback();
		m_scan_converter_callback->AddRef();
		// set output window where callback will display received ultrasound frames
//		m_scan_converter_callback->SetUsgWnd((HWND)(*this), 20, 20);

		// get IUsgScanConverterPlugin for desired ultrasound data *stream*:
		// SCAN_MODE_B, SCAN_MODE_M, SCAN_MODE_PW, SCAN_MODE_CFM, SCAN_MODE_PDI, SCAN_MODE_DPDI
		IUnknown* tmp_obj = NULL;
		IUsgfwScanConverterPlugin3* scp = NULL;
		CreateUsgControl(m_data_view, IID_IUsgfwScanConverterPlugin3, SCAN_MODE_B, 0, (void**)&tmp_obj);
		if (tmp_obj != NULL)
			scp = (IUsgfwScanConverterPlugin3*)tmp_obj;

		// get IUsgfwScanConverterPlugin4 interface
		IUnknown* tmp_unk = NULL;
		m_plugin = NULL;
		if (scp != NULL)
			scp->GetScanConverterObj(&tmp_unk);
		if (tmp_unk != NULL)
		{
			hr = tmp_unk->QueryInterface(IID_IUsgfwScanConverterPlugin4,(void**)&m_plugin);
			if (hr != S_OK)
				m_plugin = NULL;
			SAFE_RELEASE(tmp_unk);
		}

		// register COM callback object and specify method(s) to callback
		if ( (m_plugin != NULL) && (m_scan_converter_callback != NULL) )
		{
			hr = m_plugin->AddCallback(m_scan_converter_callback, USPC2_SAMPLE_INPUT, 0, 0);
			m_scan_converter_callback->SetPlugin(m_plugin);
		}

		m_data_view->put_ScanState(SCAN_STATE_STOP); // workaround
		m_data_view->put_ScanState(SCAN_STATE_FREEZE);

		return;
	} while (false);

	ReleaseUsgControls();

} // void CsampleDlg::CreateUsgControls()

//--------------------------------------------------------------------------

void CsampleDlg::ReleaseUsgControls()
{
	g_save_enabled			= 0;
	g_frames_file_number	= 0;
	g_fn_str				= "";
	if (m_data_view != NULL)
		m_data_view->put_ScanState(SCAN_STATE_STOP);
	if ( (m_plugin != NULL) && (m_scan_converter_callback != NULL) )
		m_plugin->RemoveCallback(m_scan_converter_callback); // unregister callback
	SAFE_RELEASE(m_scan_converter_callback);
	SAFE_RELEASE(m_plugin);
	SAFE_RELEASE(m_mixer_control);
	SAFE_RELEASE(m_storage);
	SAFE_RELEASE(m_cine);
	SAFE_RELEASE(m_data_view);
	//SAFE_RELEASE(m_probe);
	//SAFE_RELEASE(m_usgfw2);
} // void CsampleDlg::ReleaseUsgControls()

//--------------------------------------------------------------------------

CsampleDlg::CsampleDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CsampleDlg::IDD, pParent)
{
	m_hIcon = AfxGetApp()->LoadIcon(IDR_MAINFRAME);
}

//------------------------------

void CsampleDlg::OnClose()
{
	g_close_in_progress = true;
	if (g_conversion_in_progress == true)
	{
		g_stop_conversion = true;
		Sleep(1000);
		g_conversion_in_progress = false;
	}

	ReleaseUsgControls();

	SAFE_RELEASE(m_usgfw2);

	__super::OnClose();
}

//------------------------------

CsampleDlg::~CsampleDlg()
{
	CoUninitialize();
}

//------------------------------

void CsampleDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	DDX_Control(pDX, IDC_FILE_CTRL, m_file_ctrl);
	DDX_Control(pDX, IDC_STATUS_STATIC, m_status_ctrl);
	DDX_Control(pDX, IDC_SAVE_FRAME_FORMAT_INFO_CHECKBOX, m_save_frame_format_info_checkbox);
	DDX_Control(pDX, IDC_SAVE_SCAN_LINES_CHECKBOX, m_save_scan_lines_checkbox);
}

//------------------------------

BEGIN_MESSAGE_MAP(CsampleDlg, CDialog)
	ON_WM_PAINT()
	ON_WM_QUERYDRAGICON()
	//}}AFX_MSG_MAP
	ON_WM_CLOSE()
	ON_WM_QUERYENDSESSION()
	ON_WM_ENDSESSION()
	ON_BN_CLICKED(IDC_BROWSE_BUTTON, &CsampleDlg::OnBnClickedBrowseButton)
	ON_BN_CLICKED(IDC_OPEN_BUTTON, &CsampleDlg::OnBnClickedOpenButton)
	ON_BN_CLICKED(IDC_CONVERT_BUTTON, &CsampleDlg::OnBnClickedConvertButton)
	ON_BN_CLICKED(IDC_SAVE_FRAME_FORMAT_INFO_CHECKBOX, &CsampleDlg::OnBnClickedSaveFrameFormatInfoCheckbox)
	ON_BN_CLICKED(IDC_SAVE_SCAN_LINES_CHECKBOX, &CsampleDlg::OnBnClickedSaveScanLinesCheckbox)
	ON_BN_CLICKED(IDC_STOP_CONVERSION_BUTTON, &CsampleDlg::OnBnClickedStopConversionButton)
END_MESSAGE_MAP()

//------------------------------

BOOL CsampleDlg::OnInitDialog()
{
	CDialog::OnInitDialog();

	// Set the icon for this dialog.  The framework does this automatically
	// when the application's main window is not a dialog
	SetIcon(m_hIcon, TRUE);			// Set big icon
	SetIcon(m_hIcon, FALSE);		// Set small icon

	m_save_scan_lines_checkbox.SetCheck(BST_CHECKED);
	g_save_scan_lines = true;

	m_save_frame_format_info_checkbox.SetCheck(BST_CHECKED);
	g_save_frame_format_info = true;


	CoInitialize(NULL);

	g_save_enabled	= 0;
	g_frames_file_number	= 0;
	g_fn_str		= "";
	m_usgfw2		= NULL;
	m_data_view		= NULL;
	//m_probe			= NULL;
	m_mixer_control = NULL;
	m_storage		= NULL;
	m_cine			= NULL;
	m_plugin		= NULL;
	m_scan_converter_callback = NULL;

	CoCreateInstance(CLSID_Usgfw2, NULL, CLSCTX_INPROC_SERVER, IID_IUsgfw2,(LPVOID*) &m_usgfw2);

	//CreateUsgControls();

	return TRUE;  // return TRUE  unless you set the focus to a control
}

//------------------------------

// If you add a minimize button to your dialog, you will need the code below
//  to draw the icon.  For MFC applications using the document/view model,
//  this is automatically done for you by the framework.

void CsampleDlg::OnPaint()
{
	if (IsIconic())
	{
		CPaintDC dc(this); // device context for painting

		SendMessage(WM_ICONERASEBKGND, reinterpret_cast<WPARAM>(dc.GetSafeHdc()), 0);

		// Center icon in client rectangle
		int cxIcon = GetSystemMetrics(SM_CXICON);
		int cyIcon = GetSystemMetrics(SM_CYICON);
		CRect rect;
		GetClientRect(&rect);
		int x = (rect.Width() - cxIcon + 1) / 2;
		int y = (rect.Height() - cyIcon + 1) / 2;

		// Draw the icon
		dc.DrawIcon(x, y, m_hIcon);
	}
	else
	{
		CDialog::OnPaint();

		if (m_mixer_control != NULL)
		{
			HWND wnd;
			wnd = this->GetSafeHwnd();
			if (wnd != NULL)
			{
				HDC hdc;
				hdc = ::GetDC(wnd);
				// repaint ultrasound (important if we are in freeze)
				if ( (hdc != NULL) && (g_conversion_in_progress == false) )
					m_mixer_control->Repaint((LONG)wnd,(LONG)hdc);
				::ReleaseDC(wnd,hdc);
			}
		}

	}

}

//------------------------------

// The system calls this function to obtain the cursor to display while the user drags
// the minimized window.
HCURSOR CsampleDlg::OnQueryDragIcon()
{
	return static_cast<HCURSOR>(m_hIcon);
}

//--------------------------------------------------------------------------

//------------------------------

void CsampleDlg::OpenFile()
{
	if (m_cine != NULL)
	{
		CString fn_str;
		m_file_ctrl.GetWindowTextW(fn_str);
		if (fn_str == g_fn_str)
			return;
	}

	m_status_ctrl.SetWindowText(L"");

	HCURSOR hCurs1;
	hCurs1 = LoadCursor(NULL, IDC_WAIT);
	SetCursor(hCurs1);

	ReleaseUsgControls();
	CreateUsgControls();

	hCurs1 = LoadCursor(NULL, IDC_ARROW); 
	SetCursor(hCurs1);

	if (m_cine == NULL) return;

	hCurs1 = LoadCursor(NULL, IDC_WAIT);
	SetCursor(hCurs1);

	m_cine->put_State(CINE_STATE_STOP);
	LONGLONG pos1, pos_start, pos_end;
	m_cine->put_TimeFormat(CINE_FORMAT_FRAME);
	m_cine->SetSelection(-1,-1,SM_DESELECT);
	m_cine->SetHoldRange(-1,-1);
	m_cine->GetAvailable(&pos_start, &pos_end);
	pos1 = pos_end;
	m_cine->GetContinuousInterval(pos1, &pos_start, &pos_end);
	m_cine->put_StopPosition(pos_end);
	m_cine->put_CurrentPosition(pos_start);

	CString tmp_str;
	int tmp_num;
	tmp_num = (int)(pos_end - pos_start + 1);
	tmp_str.Format(L"Frames count: %d.", tmp_num );
	m_status_ctrl.SetWindowText(tmp_str);

	hCurs1 = LoadCursor(NULL, IDC_ARROW); 
	SetCursor(hCurs1);
}

void CsampleDlg::ConvertFile()
{
	g_conversion_in_progress = true;

	if (m_cine == NULL)
		OpenFile();

	CString fn_str;
	m_file_ctrl.GetWindowTextW(fn_str);
	if (fn_str != g_fn_str)
		OpenFile();

	if (m_cine == NULL)
	{
		g_conversion_in_progress = false;
		return;
	}

	HCURSOR hCurs1;
	hCurs1 = LoadCursor(NULL, IDC_WAIT);
	SetCursor(hCurs1);

	m_cine->put_State(CINE_STATE_STOP);
	LONGLONG pos1, pos_start, pos_end;
	m_cine->put_TimeFormat(CINE_FORMAT_FRAME);
	m_cine->SetSelection(-1,-1,SM_DESELECT);
	m_cine->SetHoldRange(-1,-1);
	m_cine->GetAvailable(&pos_start, &pos_end);
	pos1 = pos_end;
	m_cine->GetContinuousInterval(pos1, &pos_start, &pos_end);
	m_cine->put_StopPosition(pos_end);
	m_cine->put_CurrentPosition(pos_start);
	g_frames_file_number = 0;
	g_save_enabled = 0;
	g_frame_time_old = 0;
	g_frame_time_diff = 0;
	g_stop_conversion = false;
	MSG msg;
	for (LONGLONG i1=pos_start; i1<=pos_end; i1++)
	{
		g_frame_number = (int)(i1 - pos_start);
		CString tmp_str;
		CString tmp_str2;
		int tmp_num1, tmp_num2;
		tmp_num1 = (int)(i1 - pos_start + 1);
		tmp_num2 = (int)(pos_end - pos_start + 1);
		tmp_str.Format(L"Frame %d of ", tmp_num1);
		tmp_str2.Format(L"%d.", tmp_num2);
		tmp_str = tmp_str + tmp_str2;
		m_status_ctrl.SetWindowText(tmp_str);
		m_status_ctrl.Invalidate();
		Sleep(10);
		m_status_ctrl.UpdateWindow();
		Sleep(10);
		m_cine->ConvertTimeFormat(i1, CINE_FORMAT_FRAME, CINE_FORMAT_REFERENCE_TIME, &g_frame_time);
		if (i1 == pos_start)
			g_frame_time_old = g_frame_time;
		g_frame_time_diff = g_frame_time - g_frame_time_old;
		g_frame_time_old = g_frame_time;
		// enable saving only here, because GUI updates cause that frames are pushed through
		// scan-converter and callback
		g_save_enabled = 1;
		m_cine->put_CurrentPosition(i1);
		g_save_enabled = 0;
		
		if (g_close_in_progress == true)
			return;

		while (::PeekMessage(&msg, NULL, 0,0, PM_NOREMOVE))
		{
			if (g_close_in_progress == true)
				return;

			if (!AfxGetApp()->PumpMessage())
			{
				g_stop_conversion = true;
				break;
			}
		}

		if (g_stop_conversion == true)
			break;

	} // for

	g_save_enabled = 0;

	hCurs1 = LoadCursor(NULL, IDC_ARROW); 
	SetCursor(hCurs1);

	g_conversion_in_progress = false;
}

void CsampleDlg::OnBnClickedBrowseButton()
{
	if (g_conversion_in_progress == true)
		return;

	CFileDialog dlg(TRUE,NULL,NULL,OFN_PATHMUSTEXIST|OFN_FILEMUSTEXIST,L"Image Files (*.tvd;*.tpd)|*.tvd; *.tpd|TVD Files (*.tvd)|*.tvd|TPD Files (*.tpd)|*.tpd||",this);
	if (dlg.DoModal() == IDOK)
	{
		m_file_ctrl.SetWindowText(dlg.GetPathName());
		OpenFile();
	}
}

void CsampleDlg::OnBnClickedOpenButton()
{
	if (g_conversion_in_progress == true)
		return;

	OpenFile();
}

void CsampleDlg::OnBnClickedConvertButton()
{
	if (g_conversion_in_progress == true)
		return;

	ConvertFile();
}

void CsampleDlg::OnBnClickedStopConversionButton()
{
	g_stop_conversion = true;
}

void CsampleDlg::OnBnClickedSaveFrameFormatInfoCheckbox()
{
	if (m_save_frame_format_info_checkbox.GetCheck() == BST_CHECKED)
		g_save_frame_format_info = true;
	else
		g_save_frame_format_info = false;
}

void CsampleDlg::OnBnClickedSaveScanLinesCheckbox()
{
	if (m_save_scan_lines_checkbox.GetCheck() == BST_CHECKED)
		g_save_scan_lines = true;
	else
		g_save_scan_lines = false;
}

//--------------------------------------------------------------------------

BOOL CsampleDlg::OnQueryEndSession()
{
	if (!__super::OnQueryEndSession())
		return FALSE;

	return TRUE;
}

//------------------------------

void CsampleDlg::OnEndSession(BOOL bEnding)
{
	__super::OnEndSession(bEnding);

	ReleaseUsgControls();
}

//----------------------------------------------------------------
//----------------------------------------------------------------
//----------------------------------------------------------------

HRESULT __stdcall ScanConverterCallback::QueryInterface( REFIID riid, void **ppObj )
{
	if ( riid == IID_IUnknown || riid == IID_IUsgfwScanConverterPluginCB2)
	{
		*ppObj = (void *)(this); 
		AddRef();
		return S_OK;
	}
	*ppObj = NULL;
	return E_NOINTERFACE ;
}

//------------------------------

ULONG STDMETHODCALLTYPE ScanConverterCallback::AddRef( void )
{
	return InterlockedIncrement(&m_nRefCount);
}

//------------------------------

ULONG STDMETHODCALLTYPE ScanConverterCallback::Release( void )
{
	long nRefCount = 0;
	nRefCount = InterlockedDecrement(&m_nRefCount) ;
	if (nRefCount == 0)
		delete this;
	return nRefCount;
}

//------------------------------

ScanConverterCallback::ScanConverterCallback()
{
	m_nRefCount = 0;
	m_plugin = NULL;
}

//------------------------------

ScanConverterCallback::~ScanConverterCallback()
{
}

//------------------------------

void ScanConverterCallback::SetPlugin(IUsgfwScanConverterPlugin4* plugin)
{
	m_plugin = plugin;
}

//------------------------------

// Return the number of bytes used to store a scanline for the DIB.
// Rounded up to the nearest 32-bit boundary.
// ScanBytes(infoHeader.biWidth, infoHeader.biBitCount)
int ScanBytes(long w, int bpp)
{
	return (int)(((w*bpp+31)&(~31))/8);
}

#define id_b_scanning_type_unknown		-1
#define id_b_scanning_type_standard		1
#define id_b_scanning_type_steering		2
#define id_b_scanning_type_trapezoid	3
#define id_b_scanning_type_compound		4

HRESULT ScanConverterCallback::InputSampleCB(THIS_
		IMediaSample* pSample,
		LONG Param1,
		LONG Param2,
		LONG Param3,
		LONG Param4
		)
{
	if (g_close_in_progress == true)
			return S_OK;

	if ( (m_plugin == NULL) || (pSample == NULL) )
		return S_OK;

	// save scanlines

	if (g_save_enabled == 0)
		return S_OK;

	AM_MEDIA_TYPE mtIn;
	if (m_plugin->GetCurrentMediaType(PINDIR_INPUT, &mtIn) != S_OK)
		return S_OK;

	HRESULT hr;
	int Width = 0;
	int Height = 0;
	int Pitch = 0;
	LONG frame_angle = 0;
	ULONG frame_angle_num = 0;
	ULONG frame_angle_den = 0;
	int time_per_beam = 0;
	int time_per_point = 0;
	int b_scanning_type_id;
	b_scanning_type_id = id_b_scanning_type_unknown;
	USGBEAM_VECTOR* beam_vector = NULL;
	CString beam_vector_str;
	// um - micrometers
	// rad - radians
	beam_vector_str = L"beams_start_x_um_start_y_um_angle_rad_numerator_denominator";

	// standard B frame
	if (IsEqualGUID(mtIn.formattype, FORMAT_BUltrasoundInfo))
	{
		b_scanning_type_id = id_b_scanning_type_standard;

		PUSGFRAMEINFOHEADER pUsgFrameInfoHeader = NULL;
		PBULTRASOUND_INFO pFrameInfo = NULL;
		pFrameInfo = (PBULTRASOUND_INFO)mtIn.pbFormat;
		if (pFrameInfo == NULL)
		{
			FreeMediaType(mtIn);
			return S_OK;
		}
		pUsgFrameInfoHeader	= (PUSGFRAMEINFOHEADER)&pFrameInfo->UsgFrameInfoHeader;

		if (pUsgFrameInfoHeader == NULL)
		{
			FreeMediaType(mtIn);
			return S_OK;
		}

		Width	= pUsgFrameInfoHeader->PointsPerBeam;	// points per beam
		Height	= pUsgFrameInfoHeader->BeamsPerFrame;	// beams per frame
		Pitch	= pUsgFrameInfoHeader->Pitch;	// offset in bytes between two sequential beams
		time_per_beam	= pUsgFrameInfoHeader->TimePerBeam;
		time_per_point	= pUsgFrameInfoHeader->TimePerPoint;

		beam_vector = pFrameInfo->UsgBeamVector;
	}
	// steering, trapezoid or compound
	else if (IsEqualGUID(mtIn.formattype,FORMAT_UltrasoundInfo2D))
	{
		ULTRASOUND2D_INFO* pUltrasoundHdr;
		pUltrasoundHdr = (ULTRASOUND2D_INFO*)mtIn.pbFormat;
		if (pUltrasoundHdr == NULL)
		{
			FreeMediaType(mtIn);
			return S_OK;
		}

		Width	= pUltrasoundHdr->UsgFrameInfoHeader.PointsPerBeam;	// points per beam
		Height	= pUltrasoundHdr->UsgFrameInfoHeader.BeamsPerFrame;	// beams per frame
		Pitch	= pUltrasoundHdr->UsgFrameInfoHeader.Pitch;	// offset in bytes between two sequential beams
		time_per_beam	= pUltrasoundHdr->UsgFrameInfoHeader.TimePerBeam;
		time_per_point	= pUltrasoundHdr->UsgFrameInfoHeader.TimePerPoint;

		beam_vector = pUltrasoundHdr->UsgBeamVector;

		if (IsEqualGUID(mtIn.subtype, MEDIASUBTYPE_Steering2D))
		{
			b_scanning_type_id = id_b_scanning_type_steering;
		}
		else if (IsEqualGUID(mtIn.subtype, MEDIASUBTYPE_VirtualConvex2D))
		{
			b_scanning_type_id = id_b_scanning_type_trapezoid;
		}
		else if (IsEqualGUID(mtIn.subtype, MEDIASUBTYPE_Compound2D))
		{
			b_scanning_type_id = id_b_scanning_type_compound;

			IMediaSample2 *pSampleIn2 = NULL;
			if (SUCCEEDED(pSample->QueryInterface(IID_IMediaSample2,	(void **)&pSampleIn2)))
			{
				AM_SAMPLE2_PROPERTIES Props2;
				if (SUCCEEDED(pSampleIn2->GetProperties(FIELD_OFFSET(AM_SAMPLE2_PROPERTIES, tStart), (PBYTE)&Props2)))
				{
					int tmp_idx;
					// extract bits D7:D4
					tmp_idx = (Props2.dwTypeSpecificFlags & 0xF0) >> 4;
					frame_angle = beam_vector[Height * tmp_idx].Direction.value;
					frame_angle_num = beam_vector[Height * tmp_idx].Direction.ScaleFactor.Numerator;
					frame_angle_den = beam_vector[Height * tmp_idx].Direction.ScaleFactor.Denominator;

					// get data from beam vector for compound scanning type
					// the difference from other scanning types is that to callback we receive
					// only subframe (not full frame), but beam_vector contains data for all subframes
					if (g_save_frame_format_info == true)
					{
						if (beam_vector != NULL)
						{
							CString str3;
							for (int i1=0; i1<Height; i1++)
							{
								str3.Format(L";%ld;%ld;%ld;%lu;%lu", beam_vector[Height * tmp_idx + i1].StartPoint.x, beam_vector[Height * tmp_idx + i1].StartPoint.y, beam_vector[Height * tmp_idx + i1].Direction.value, beam_vector[Height * tmp_idx + i1].Direction.ScaleFactor.Numerator, beam_vector[Height * tmp_idx + i1].Direction.ScaleFactor.Denominator);
								beam_vector_str += str3;
							}
						} // if (beam_vector != NULL)
					}

				}
				pSampleIn2->Release();
			}
		} // else if (IsEqualGUID(mtIn.subtype,MEDIASUBTYPE_Compound2D))
	} // else if (IsEqualGUID(mtIn.formattype,FORMAT_UltrasoundInfo2D))

	if ( (Width == 0) && (Height == 0) && (Pitch == 0) )
	{
		FreeMediaType(mtIn);
		return S_OK;
	}

	FreeMediaType(mtIn);

	PBYTE pBufferIn = NULL;
	int nInBufferLen;
	nInBufferLen = pSample->GetSize();
	hr = pSample->GetPointer(&pBufferIn);
	if (FAILED(hr))
	{
		return S_OK;
	}

	PBYTE ln_buf;
	ln_buf = pBufferIn;

	if (g_save_frame_format_info == true)
	{
		// save frame format information
		CString fn;
		CString ext1;
		ext1.Format(L".%04d.txt", g_frames_file_number);
		fn = g_fn_str + ext1;

		CStdioFile file;
		if (file.Open(fn, CFile::modeCreate | CFile::modeWrite | CFile::typeText) == FALSE)
			return S_OK;

		CString str;
		CString str2;
		str = L"";

		if (b_scanning_type_id == id_b_scanning_type_standard)
			str2 = L"b_scanning_type;standard\r\n";
		else if (b_scanning_type_id == id_b_scanning_type_steering)
			str2 = L"b_scanning_type;steering\r\n";
		else if (b_scanning_type_id == id_b_scanning_type_trapezoid)
			str2 = L"b_scanning_type;trapezoid\r\n";
		else if (b_scanning_type_id == id_b_scanning_type_compound)
			str2 = L"b_scanning_type;compound\r\n";
		else
			str2 = L"b_scanning_type;unknown\r\n";
		str += str2;

		str2.Format(L"frame_number;%d\r\n", g_frame_number);
		str += str2;
		// 100 ns - 100 nanoseconds
		str2.Format(L"frame_reference_time_100ns;%llu\r\n", g_frame_time);
		str += str2;
		str2.Format(L"frame_reference_time_difference_100ns;%llu\r\n", g_frame_time_diff);
		str += str2;
		str2.Format(L"frames_file_number;%d\r\n", g_frames_file_number);
		str += str2;

		str2.Format(L"input_points_per_beam;%d\r\n", Width);
		str += str2;
		str2.Format(L"input_beams_per_frame;%d\r\n", Height);
		str += str2;
		str2.Format(L"input_pitch_bytes;%d\r\n", Pitch);
		str += str2;
		str2.Format(L"input_buffer_length_bytes;%d\r\n", nInBufferLen);
		str += str2;
		// us - microseconds
		str2.Format(L"time_per_beam_us;%d\r\n", time_per_beam);
		str += str2;
		// ns - nanoseconds
		str2.Format(L"time_per_point_ns;%d\r\n", time_per_point);
		str += str2;

		if (b_scanning_type_id == id_b_scanning_type_compound)
		{
			str2.Format(L"compound_subframe_angle_rad_numerator_denominator;%ld;%lu;%lu\r\n", frame_angle, frame_angle_num, frame_angle_den);
			str += str2;
		}

		if (beam_vector != NULL)
		{
			if (b_scanning_type_id != id_b_scanning_type_compound)
			{
				// get data from beam vector for non-compound scanning types (for compund see code above)
				for (int i1=0; i1<Height; i1++)
				{
					CString str3;
					str3.Format(L";%ld;%ld;%ld;%lu;%lu", beam_vector[i1].StartPoint.x, beam_vector[i1].StartPoint.y, beam_vector[i1].Direction.value, beam_vector[i1].Direction.ScaleFactor.Numerator, beam_vector[i1].Direction.ScaleFactor.Denominator);
					beam_vector_str += str3;
				}
			}
			str += beam_vector_str;
			str += "\r\n";
		} // if (beam_vector != NULL)

		file.WriteString(str);
		file.Close();

		if (g_close_in_progress == true)
			return S_OK;

	} // if (g_save_frame_format_info == true)

	if (g_save_scan_lines == true)
	{
		// create file for each scanline
		for (int i1=0; i1<Height; i1++)
		{
			CString fn;
			CString ext1;
			ext1.Format(L".%04d.%03d.bin", g_frames_file_number, i1);
			fn = g_fn_str + ext1;

			CFile file;
			if (file.Open(fn, CFile::modeCreate | CFile::modeCreate | CFile::modeWrite | CFile::typeBinary) == FALSE)
				return S_OK;

			file.Write(ln_buf, Width);
			file.Close();

			if (g_stop_conversion == true)
				return S_OK;
			if (g_close_in_progress == true)
				return S_OK;

			ln_buf += Pitch;
		}
	} // if (g_save_scan_lines == true)

	g_frames_file_number++;

	return S_OK;
} // HRESULT ScanConverterCallback::BufferCB

//----------------------------------------------------------------

